//
// Pool manager client declaration definition.
//
// Copyright (c) 2019 NVIDIA Corporation. All rights reserved.
//
// NVIDIA Corporation and its licensors retain all intellectual property and
// proprietary rights in and to this software, related documentation and any
// modifications thereto. Any use, reproduction, disclosure or distribution
// of this software and related documentation without an express license
// agreement from NVIDIA Corporation is strictly prohibited.

#include "packet_pool_manager.h"

namespace NvScistreamSample
{
    PacketPoolManager::PacketPoolManager(NvSciStreamBlock blockHandle):
        handle(blockHandle),
        numProdElements(0U),
        numProdElementsRecv(0U),
        prodElementAttrs(nullptr),
        numConsElements(0U),
        numConsElementsRecv(0U),
        consElementAttrs(nullptr),
        numElements(0U),
        elementAttrs(nullptr),
        numPackets(NUM_PACKETS),
        numPacketsCreated(0U),
        numProdPacketStatusRecv(0U),
        numConsPacketStatusRecv(0U),
        numProdElementStatusRecv(0U),
        numConsElementStatusRecv(0U)
    {
    }

    PacketPoolManager::~PacketPoolManager(void)
    {
        freeElementAttrs(prodElementAttrs, numProdElements);
        freeElementAttrs(consElementAttrs, numConsElements);
        freeElementAttrs(elementAttrs, numElements);

        for (uint32_t i = 0U; i < numPackets; i++) {
            for (uint32_t j = 0; j < numElements; j++) {
                NvSciBufObjFree(packets[i].buffers[j]);
            }
        }
    }

    // Receive producer's packet capabilities and consumer's packet
    //  requirements.
    void PacketPoolManager::recvPacketAttrs(NvSciStreamEvent &event)
    {
        uint32_t index;

        switch (event.type) {
            case NvSciStreamEventType_PacketElementCountProducer:
                numProdElements = event.count;
                LOG_DEBUG("Pool Receives number of elements per packet from producer: " << event.count << ".\n");

                // Allocate space for producer packet element attr list
                prodElementAttrs = static_cast<NvSciStreamElementAttr*>(
                    calloc(numProdElements, sizeof(NvSciStreamElementAttr)));

                break;

            case NvSciStreamEventType_PacketElementCountConsumer:
                numConsElements = event.count;
                LOG_DEBUG("Pool Receives number of elements per packet from consumer: " << event.count << ".\n");

                // Allocate space for producer packet element attr list
                consElementAttrs = static_cast<NvSciStreamElementAttr*>(
                    calloc(numConsElements, sizeof(NvSciStreamElementAttr)));

                break;

            case NvSciStreamEventType_PacketAttrProducer:
                if (numProdElements == 0) {
                    LOG_ERR_EXIT("\nInvalid number of producer elements");
                }

                index = event.packetAttr.index;
                copyElementAttr(prodElementAttrs[index], event.packetAttr, false);
                LOG_DEBUG("Pool Receives packet capabilities of element " << index << " from producer.\n");
                numProdElementsRecv++;
                break;

            case NvSciStreamEventType_PacketAttrConsumer:
                if (numConsElements == 0) {
                    LOG_ERR_EXIT("\nInvalid number of consumer elements");
                }

                index = event.packetAttr.index;
                copyElementAttr(consElementAttrs[index], event.packetAttr, false);
                LOG_DEBUG("Pool Receives packet requirements of element " << index << " from consumer.\n");
                numConsElementsRecv++;
                break;

            default:
                break;
        }
    }

    bool PacketPoolManager::recvAllPacketAttrs(void)
    {
        return ((numProdElements > 0) &&
                (numProdElementsRecv >= numProdElements) &&
                (numConsElements > 0) &&
                (numConsElementsRecv >= numConsElements));
    }

    // Determine the number of elements per packet. It chooses the bigger
    //  one from the producer's and consumer's numbers of elements.
    void PacketPoolManager::determineNumPacketElements(void)
    {
        int32_t maxNumElements = 0U;
        CHECK_NVSCIERR(NvSciStreamAttributeQuery(
                NvSciStreamQueryableAttrib_MaxElements,
                &maxNumElements));
        LOG_DEBUG("Query max number of elements per packet allowed: " << maxNumElements << ".\n");

        // Pool determines the number of packet elements
        if (numProdElements > numElements) {
            numElements = numProdElements;
        }
        if (numConsElements > numElements) {
            numElements = numConsElements;
        }

        // Ensure the number of sync objects is supported.
        if (numElements > static_cast<uint32_t>(maxNumElements)) {
            LOG_ERR_EXIT("\nInvalid num of elements.\n");
        }

        // The default number of elements per packet is 1. If the number of
        // elements per packet is not 1, need to notify the other end.
        if (numElements != 1U) {
            CHECK_NVSCIERR(NvSciStreamBlockPacketElementCount(handle, numElements));
            LOG_DEBUG("Send the determined number of packet elements " << numElements << ".\n");
        }

        LOG_DEBUG("Determine the number of elements per packet: " << numElements << ".\n");
    }

    // Reconcile the producer's and consumer's attribute lists for each
    //  packet element.
    void PacketPoolManager::reconcilePacketAttrs(void)
    {
        // Allocate space for packet element attributes
        elementAttrs = static_cast<NvSciStreamElementAttr*>(
            calloc(numElements, sizeof(NvSciStreamElementAttr)));

        // Pool determines the attribute list for each packet element
        for (uint32_t i = 0U; i < numElements; i++) {
            elementAttrs[i].index = i;
            if (i < numProdElements && i < numConsElements) {
                if (prodElementAttrs[i].index != consElementAttrs[i].index ||
                    prodElementAttrs[i].mode != consElementAttrs[i].mode ||
                    prodElementAttrs[i].type != consElementAttrs[i].type) {
                    LOG_ERR_EXIT("\nAttribute list mismatch");
                }
                elementAttrs[i].mode = prodElementAttrs[i].mode;
                elementAttrs[i].type = prodElementAttrs[i].type;
            }
            else if (i < numProdElements) {
                elementAttrs[i].mode = prodElementAttrs[i].mode;
                elementAttrs[i].type = prodElementAttrs[i].type;
            }
            else if (i < numConsElements) {
                elementAttrs[i].mode = consElementAttrs[i].mode;
                elementAttrs[i].type = consElementAttrs[i].type;
            }

            // Reconcile buffer attribute list
            NvSciBufAttrList attrlists[2] = { nullptr };
            NvSciBufAttrList conflictlist = nullptr;
            if (i < numProdElements) {
                attrlists[0] = prodElementAttrs[i].bufAttr;
            }
            if (i < numConsElements) {
                attrlists[1] = consElementAttrs[i].bufAttr;
            }
            CHECK_NVSCIERR(NvSciBufAttrListReconcile(
                    attrlists,
                    2,
                    &elementAttrs[i].bufAttr,
                    &conflictlist));
            LOG_DEBUG("Reconcile producer's and consumer's attribute lists of element " << i << ".\n");

            // Free the conflict attribute lists
            if (conflictlist != nullptr) {
                NvSciBufAttrListFree(conflictlist);
                conflictlist = nullptr;
            }
        }
    }
    // Send the determined packet layout to both producer and consumer.
    void PacketPoolManager::sendReconciledPacketAttrs(void)
    {
        for (uint32_t i = 0U; i < numElements; i++) {
            NvSciError err = NvSciStreamBlockPacketAttr(handle, &elementAttrs[i]);
            LOG_DEBUG("Send the reconciled attribute list of element " << i << " to producer and consumer.\n");
            CHECK_NVSCIERR(err);
        }
    }

    // Create a buffer object for each element.
    void PacketPoolManager::allocBuffers(NvSciStreamCookie &cookie)
    {
        // Assign cookie for the new packet
        cookie = assignPacketCookie();

        // Get the slot by pool's cookie and allocate space for all elements
        // (buffers) in this packet.
        Packet *packet = getPacketByCookie(cookie);
        if (packet == nullptr) {
            LOG_ERR_EXIT("\nInvalid packet for cookie.\n");
        }

        packet->cookie = cookie;
        packet->buffers = static_cast<NvSciBufObj*>(
            calloc(numElements, sizeof(NvSciBufObj)));

        // Create buffer object for each element.
        for (uint32_t i = 0U; i < numElements; i++) {
            CHECK_NVSCIERR(NvSciBufObjAlloc(
                    elementAttrs[i].bufAttr,
                    &packet->buffers[i]));
            LOG_DEBUG("Create buffer object of element " << i << " with the reconciled attribute list.\n");
        }

        numPacketsCreated++;
    }

    // Send a new packet and all the packet elements to the producer
    //  and consumer.
    void PacketPoolManager::sendPacket(NvSciStreamCookie cookie)
    {
        // Send pool's packet cookie to the NvSciStream, and save the packet handle
        // returned from the NvSciStream.
        Packet *packet = getPacketByCookie(cookie);
        if (packet == nullptr) {
            LOG_ERR_EXIT("\nInvalid packet for cookie.\n");
        }

        CHECK_NVSCIERR(NvSciStreamPoolPacketCreate(
                handle,
                packet->cookie,
                &packet->handle));
        LOG_DEBUG("Assign pool cookie " << packet->cookie << " to the packet, and\n"
                "\t\t\tsend this new packet (handle = " << std::hex << packet->handle << ") to producer and consumer.\n");

        // Send each element (buffer) one by one.
        for (uint32_t i = 0U; i < numElements; i++) {
            NvSciStreamElementDesc desc;
            desc.index = i;
            desc.handle = packet->handle;
            desc.buffer = packet->buffers[i];

            CHECK_NVSCIERR(NvSciStreamPoolPacketInsertBuffer(
                    handle,
                    &desc));
            LOG_DEBUG("Send buffer object of element " << i << " to producer and consumer.\n");
        }
    }

    void PacketPoolManager::reconcileAndMapPackets(void)
    {
        determineNumPacketElements();
        reconcilePacketAttrs();
        sendReconciledPacketAttrs();

        // Pool allocates buffers and sends packets to producer
        // + consumer
        for (uint32_t i = 0U; i < NUM_PACKETS; i++) {
            NvSciStreamCookie poolCookie = 0U;
            LOG_DEBUG("Pool creates a new packet " << i << ":\n");
            allocBuffers(poolCookie);
            sendPacket(poolCookie);
        }
    }

    // Receive the packet mapping status from producer and consumer.
    //  Buffer registration is finished.
    void PacketPoolManager::recvPacketStatus(NvSciStreamEvent &event)
    {
        // process packet status event
        switch(event.type) {
            case NvSciStreamEventType_PacketStatusProducer:
                numProdPacketStatusRecv++;
                if (numProdPacketStatusRecv > NUM_PACKETS) {
                    LOG_ERR_EXIT("\nReceived excess packet status from producer: " << numProdPacketStatusRecv << "\n");
                }
                LOG_DEBUG("Received PacketStatus (cookie = " << event.packetCookie << ")"
                            " from producer, total received = " << numProdPacketStatusRecv << ".\n");
                break;
            case NvSciStreamEventType_PacketStatusConsumer:
                numConsPacketStatusRecv++;
                if (numConsPacketStatusRecv > NUM_PACKETS) {
                    LOG_ERR_EXIT("\nReceived excess packet status from producer: " << numProdPacketStatusRecv << "\n");
                }
                LOG_DEBUG("Received PacketStatus (cookie = " << event.packetCookie << ")"
                            " from consumer, total received = " << numConsPacketStatusRecv << ".\n");
                break;
            case NvSciStreamEventType_ElementStatusProducer:
                numProdElementStatusRecv++;
                if (numProdElementStatusRecv > (NUM_PACKETS * numElements)) {
                    LOG_ERR_EXIT("\nReceived excess element status from producer: " << numProdElementStatusRecv << "\n");
                }
                LOG_DEBUG("Received ElementStatus (cookie = " << event.packetCookie << ", idx = " << event.index << ")"
                            " from producer, total received = " << numProdElementStatusRecv << ".\n");
                break;
            case NvSciStreamEventType_ElementStatusConsumer:
                numConsElementStatusRecv++;
                if (numConsElementStatusRecv > (NUM_PACKETS * numElements)) {
                    LOG_ERR_EXIT("\nReceived excess element status from consumer: " << numConsElementStatusRecv << "\n");
                }
                LOG_DEBUG("Received ElementStatus (cookie = " << event.packetCookie << ", idx = " << event.index << ")"
                            " from consumer, total received = " << numConsElementStatusRecv << ".\n");
                break;
            default:
                break;
        }
    }

    bool PacketPoolManager::recvAllPacketStatus(void)
    {
        const uint32_t expectedElementCount = NUM_PACKETS * numElements;
        return ((numProdPacketStatusRecv >= NUM_PACKETS) &&
                (numProdElementStatusRecv >= expectedElementCount) &&
                (numConsPacketStatusRecv >= NUM_PACKETS) &&
                (numConsElementStatusRecv >= expectedElementCount));
    }
}
